#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

from typing import Any, List

from elasticsearch import Elasticsearch
from pytest import raises

from elasticsearch_dsl import connections, serializer


class DummyElasticsearch:
    def __init__(self, *args: Any, hosts: List[str], **kwargs: Any):
        self.hosts = hosts


def test_default_connection_is_returned_by_default() -> None:
    c = connections.Connections[object](elasticsearch_class=object)

    con, con2 = object(), object()
    c.add_connection("default", con)

    c.add_connection("not-default", con2)

    assert c.get_connection() is con


def test_get_connection_created_connection_if_needed() -> None:
    c = connections.Connections[DummyElasticsearch](
        elasticsearch_class=DummyElasticsearch
    )
    c.configure(
        default={"hosts": ["https://es.com:9200"]},
        local={"hosts": ["https://localhost:9200"]},
    )

    default = c.get_connection()
    local = c.get_connection("local")

    assert isinstance(default, DummyElasticsearch)
    assert isinstance(local, DummyElasticsearch)

    assert default.hosts == ["https://es.com:9200"]
    assert local.hosts == ["https://localhost:9200"]


def test_configure_preserves_unchanged_connections() -> None:
    c = connections.Connections[DummyElasticsearch](
        elasticsearch_class=DummyElasticsearch
    )

    c.configure(
        default={"hosts": ["https://es.com:9200"]},
        local={"hosts": ["https://localhost:9200"]},
    )
    default = c.get_connection()
    local = c.get_connection("local")

    c.configure(
        default={"hosts": ["https://not-es.com:9200"]},
        local={"hosts": ["https://localhost:9200"]},
    )
    new_default = c.get_connection()
    new_local = c.get_connection("local")

    assert new_local is local
    assert new_default is not default


def test_remove_connection_removes_both_conn_and_conf() -> None:
    c = connections.Connections[object](elasticsearch_class=DummyElasticsearch)

    c.configure(
        default={"hosts": ["https://es.com:9200"]},
        local={"hosts": ["https://localhost:9200"]},
    )
    c.add_connection("local2", object())

    c.remove_connection("default")
    c.get_connection("local2")
    c.remove_connection("local2")

    with raises(Exception):
        c.get_connection("local2")
        c.get_connection("default")


def test_create_connection_constructs_client() -> None:
    c = connections.Connections[DummyElasticsearch](
        elasticsearch_class=DummyElasticsearch
    )
    c.create_connection("testing", hosts=["https://es.com:9200"])

    con = c.get_connection("testing")
    assert con.hosts == ["https://es.com:9200"]


def test_create_connection_adds_our_serializer() -> None:
    c = connections.Connections[Elasticsearch](elasticsearch_class=Elasticsearch)
    c.create_connection("testing", hosts=["https://es.com:9200"])

    c_serializers = c.get_connection("testing").transport.serializers
    assert c_serializers.serializers["application/json"] is serializer.serializer


def test_connection_has_correct_user_agent() -> None:
    c = connections.Connections[Elasticsearch](elasticsearch_class=Elasticsearch)

    c.create_connection("testing", hosts=["https://es.com:9200"])
    assert (
        c.get_connection("testing")
        ._headers["user-agent"]
        .startswith("elasticsearch-dsl-py/")
    )

    my_client = Elasticsearch(hosts=["http://localhost:9200"])
    my_client = my_client.options(headers={"user-agent": "my-user-agent/1.0"})
    c.add_connection("default", my_client)
    assert c.get_connection()._headers["user-agent"].startswith("elasticsearch-dsl-py/")

    my_client = Elasticsearch(hosts=["http://localhost:9200"])
    assert (
        c.get_connection(my_client)
        ._headers["user-agent"]
        .startswith("elasticsearch-dsl-py/")
    )

    not_a_client = object()
    assert c.get_connection(not_a_client) == not_a_client  # type: ignore[arg-type]



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import pickle
from typing import Any, Dict, Tuple

from pytest import raises

from elasticsearch_dsl import Q, serializer, utils


def test_attrdict_pickle() -> None:
    ad: utils.AttrDict[str] = utils.AttrDict({})

    pickled_ad = pickle.dumps(ad)
    assert ad == pickle.loads(pickled_ad)


def test_attrlist_pickle() -> None:
    al = utils.AttrList[Any]([])

    pickled_al = pickle.dumps(al)
    assert al == pickle.loads(pickled_al)


def test_attrlist_slice() -> None:
    class MyAttrDict(utils.AttrDict[str]):
        pass

    l = utils.AttrList[Any]([{}, {}], obj_wrapper=MyAttrDict)
    assert isinstance(l[:][0], MyAttrDict)


def test_attrlist_with_type_argument() -> None:
    a = utils.AttrList[str](["a", "b"])
    assert list(a) == ["a", "b"]


def test_attrdict_keys_items() -> None:
    a = utils.AttrDict({"a": {"b": 42, "c": 47}, "d": "e"})
    assert list(a.keys()) == ["a", "d"]
    assert list(a.items()) == [("a", {"b": 42, "c": 47}), ("d", "e")]


def test_attrdict_with_type_argument() -> None:
    a = utils.AttrDict[str]({"a": "b"})
    assert list(a.keys()) == ["a"]
    assert list(a.items()) == [("a", "b")]


def test_merge() -> None:
    a: utils.AttrDict[Any] = utils.AttrDict({"a": {"b": 42, "c": 47}})
    b = {"a": {"b": 123, "d": -12}, "e": [1, 2, 3]}

    utils.merge(a, b)

    assert a == {"a": {"b": 123, "c": 47, "d": -12}, "e": [1, 2, 3]}


def test_merge_conflict() -> None:
    data: Tuple[Dict[str, Any], ...] = (
        {"a": 42},
        {"a": {"b": 47}},
    )
    for d in data:
        utils.merge({"a": {"b": 42}}, d)
        with raises(ValueError):
            utils.merge({"a": {"b": 42}}, d, True)


def test_attrdict_bool() -> None:
    d: utils.AttrDict[str] = utils.AttrDict({})

    assert not d
    d.title = "Title"
    assert d


def test_attrlist_items_get_wrapped_during_iteration() -> None:
    al = utils.AttrList([1, object(), [1], {}])

    l = list(iter(al))

    assert isinstance(l[2], utils.AttrList)
    assert isinstance(l[3], utils.AttrDict)


def test_serializer_deals_with_Attr_versions() -> None:
    d = utils.AttrDict({"key": utils.AttrList([1, 2, 3])})

    assert serializer.serializer.dumps(d) == serializer.serializer.dumps(
        {"key": [1, 2, 3]}
    )


def test_serializer_deals_with_objects_with_to_dict() -> None:
    class MyClass:
        def to_dict(self) -> int:
            return 42

    assert serializer.serializer.dumps(MyClass()) == b"42"


def test_recursive_to_dict() -> None:
    assert utils.recursive_to_dict({"k": [1, (1.0, {"v": Q("match", key="val")})]}) == {
        "k": [1, (1.0, {"v": {"match": {"key": "val"}}})]
    }


def test_attrlist_to_list() -> None:
    l = utils.AttrList[Any]([{}, {}]).to_list()
    assert isinstance(l, list)
    assert l == [{}, {}]


def test_attrdict_with_reserved_keyword() -> None:
    d = utils.AttrDict({"from": 10, "size": 20})
    assert d.from_ == 10
    assert d.size == 20
    d = utils.AttrDict({})
    d.from_ = 10
    assert {"from": 10} == d.to_dict()



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)

from ..utils import _R, AttrDict, AttrList, _wrap
from .hit import Hit, HitMeta

if TYPE_CHECKING:
    from .. import types
    from ..aggs import Agg
    from ..faceted_search_base import FacetedSearchBase
    from ..search_base import Request, SearchBase
    from ..update_by_query_base import UpdateByQueryBase

__all__ = [
    "Response",
    "AggResponse",
    "UpdateByQueryResponse",
    "Hit",
    "HitMeta",
    "AggregateResponseType",
]


class Response(AttrDict[Any], Generic[_R]):
    """An Elasticsearch search response.

    :arg took: (required)
    :arg timed_out: (required)
    :arg _shards: (required)
    :arg hits: search results
    :arg aggregations: aggregation results
    :arg _clusters:
    :arg fields:
    :arg max_score:
    :arg num_reduce_phases:
    :arg profile:
    :arg pit_id:
    :arg _scroll_id:
    :arg suggest:
    :arg terminated_early:
    """

    _search: "SearchBase[_R]"
    _faceted_search: "FacetedSearchBase[_R]"
    _doc_class: Optional[_R]
    _hits: List[_R]

    took: int
    timed_out: bool
    _shards: "types.ShardStatistics"
    _clusters: "types.ClusterStatistics"
    fields: Mapping[str, Any]
    max_score: float
    num_reduce_phases: int
    profile: "types.Profile"
    pit_id: str
    _scroll_id: str
    suggest: Mapping[
        str,
        Sequence[
            Union["types.CompletionSuggest", "types.PhraseSuggest", "types.TermSuggest"]
        ],
    ]
    terminated_early: bool

    def __init__(
        self,
        search: "Request[_R]",
        response: Dict[str, Any],
        doc_class: Optional[_R] = None,
    ):
        super(AttrDict, self).__setattr__("_search", search)
        super(AttrDict, self).__setattr__("_doc_class", doc_class)
        super().__init__(response)

    def __iter__(self) -> Iterator[_R]:  # type: ignore[override]
        return iter(self.hits)

    def __getitem__(self, key: Union[slice, int, str]) -> Any:
        if isinstance(key, (slice, int)):
            # for slicing etc
            return self.hits[key]
        return super().__getitem__(key)

    def __nonzero__(self) -> bool:
        return bool(self.hits)

    __bool__ = __nonzero__

    def __repr__(self) -> str:
        return "<Response: %r>" % (self.hits or self.aggregations)

    def __len__(self) -> int:
        return len(self.hits)

    def __getstate__(self) -> Tuple[Dict[str, Any], "Request[_R]", Optional[_R]]:  # type: ignore[override]
        return self._d_, self._search, self._doc_class

    def __setstate__(
        self, state: Tuple[Dict[str, Any], "Request[_R]", Optional[_R]]  # type: ignore[override]
    ) -> None:
        super(AttrDict, self).__setattr__("_d_", state[0])
        super(AttrDict, self).__setattr__("_search", state[1])
        super(AttrDict, self).__setattr__("_doc_class", state[2])

    def success(self) -> bool:
        return self._shards.total == self._shards.successful and not self.timed_out

    @property
    def hits(self) -> List[_R]:
        if not hasattr(self, "_hits"):
            h = cast(AttrDict[Any], self._d_["hits"])

            try:
                hits = AttrList(list(map(self._search._get_result, h["hits"])))
            except AttributeError as e:
                # avoid raising AttributeError since it will be hidden by the property
                raise TypeError("Could not parse hits.", e)

            # avoid assigning _hits into self._d_
            super(AttrDict, self).__setattr__("_hits", hits)
            for k in h:
                setattr(self._hits, k, _wrap(h[k]))
        return self._hits

    @property
    def aggregations(self) -> "AggResponse[_R]":
        return self.aggs

    @property
    def aggs(self) -> "AggResponse[_R]":
        if not hasattr(self, "_aggs"):
            aggs = AggResponse[_R](
                cast("Agg[_R]", self._search.aggs),
                self._search,
                cast(Dict[str, Any], self._d_.get("aggregations", {})),
            )

            # avoid assigning _aggs into self._d_
            super(AttrDict, self).__setattr__("_aggs", aggs)
        return cast("AggResponse[_R]", self._aggs)

    def search_after(self) -> "SearchBase[_R]":
        """
        Return a ``Search`` instance that retrieves the next page of results.

        This method provides an easy way to paginate a long list of results using
        the ``search_after`` option. For example::

            page_size = 20
            s = Search()[:page_size].sort("date")

            while True:
                # get a page of results
                r = await s.execute()

                # do something with this page of results

                # exit the loop if we reached the end
                if len(r.hits) < page_size:
                    break

                # get a search object with the next page of results
                s = r.search_after()

        Note that the ``search_after`` option requires the search to have an
        explicit ``sort`` order.
        """
        if len(self.hits) == 0:
            raise ValueError("Cannot use search_after when there are no search results")
        if not hasattr(self.hits[-1].meta, "sort"):  # type: ignore
            raise ValueError("Cannot use search_after when results are not sorted")
        return self._search.extra(search_after=self.hits[-1].meta.sort)  # type: ignore


AggregateResponseType = Union[
    "types.CardinalityAggregate",
    "types.HdrPercentilesAggregate",
    "types.HdrPercentileRanksAggregate",
    "types.TDigestPercentilesAggregate",
    "types.TDigestPercentileRanksAggregate",
    "types.PercentilesBucketAggregate",
    "types.MedianAbsoluteDeviationAggregate",
    "types.MinAggregate",
    "types.MaxAggregate",
    "types.SumAggregate",
    "types.AvgAggregate",
    "types.WeightedAvgAggregate",
    "types.ValueCountAggregate",
    "types.SimpleValueAggregate",
    "types.DerivativeAggregate",
    "types.BucketMetricValueAggregate",
    "types.StatsAggregate",
    "types.StatsBucketAggregate",
    "types.ExtendedStatsAggregate",
    "types.ExtendedStatsBucketAggregate",
    "types.GeoBoundsAggregate",
    "types.GeoCentroidAggregate",
    "types.HistogramAggregate",
    "types.DateHistogramAggregate",
    "types.AutoDateHistogramAggregate",
    "types.VariableWidthHistogramAggregate",
    "types.StringTermsAggregate",
    "types.LongTermsAggregate",
    "types.DoubleTermsAggregate",
    "types.UnmappedTermsAggregate",
    "types.LongRareTermsAggregate",
    "types.StringRareTermsAggregate",
    "types.UnmappedRareTermsAggregate",
    "types.MultiTermsAggregate",
    "types.MissingAggregate",
    "types.NestedAggregate",
    "types.ReverseNestedAggregate",
    "types.GlobalAggregate",
    "types.FilterAggregate",
    "types.ChildrenAggregate",
    "types.ParentAggregate",
    "types.SamplerAggregate",
    "types.UnmappedSamplerAggregate",
    "types.GeoHashGridAggregate",
    "types.GeoTileGridAggregate",
    "types.GeoHexGridAggregate",
    "types.RangeAggregate",
    "types.DateRangeAggregate",
    "types.GeoDistanceAggregate",
    "types.IpRangeAggregate",
    "types.IpPrefixAggregate",
    "types.FiltersAggregate",
    "types.AdjacencyMatrixAggregate",
    "types.SignificantLongTermsAggregate",
    "types.SignificantStringTermsAggregate",
    "types.UnmappedSignificantTermsAggregate",
    "types.CompositeAggregate",
    "types.FrequentItemSetsAggregate",
    "types.TimeSeriesAggregate",
    "types.ScriptedMetricAggregate",
    "types.TopHitsAggregate",
    "types.InferenceAggregate",
    "types.StringStatsAggregate",
    "types.BoxPlotAggregate",
    "types.TopMetricsAggregate",
    "types.TTestAggregate",
    "types.RateAggregate",
    "types.CumulativeCardinalityAggregate",
    "types.MatrixStatsAggregate",
    "types.GeoLineAggregate",
]


class AggResponse(AttrDict[Any], Generic[_R]):
    """An Elasticsearch aggregation response."""

    _meta: Dict[str, Any]

    def __init__(self, aggs: "Agg[_R]", search: "Request[_R]", data: Dict[str, Any]):
        super(AttrDict, self).__setattr__("_meta", {"search": search, "aggs": aggs})
        super().__init__(data)

    def __getitem__(self, attr_name: str) -> AggregateResponseType:
        if attr_name in self._meta["aggs"]:
            # don't do self._meta['aggs'][attr_name] to avoid copying
            agg = self._meta["aggs"].aggs[attr_name]
            return cast(
                AggregateResponseType,
                agg.result(self._meta["search"], self._d_[attr_name]),
            )
        return super().__getitem__(attr_name)  # type: ignore

    def __iter__(self) -> Iterator[AggregateResponseType]:  # type: ignore[override]
        for name in self._meta["aggs"]:
            yield self[name]


class UpdateByQueryResponse(AttrDict[Any], Generic[_R]):
    """An Elasticsearch update by query response.

    :arg batches:
    :arg failures:
    :arg noops:
    :arg deleted:
    :arg requests_per_second:
    :arg retries:
    :arg task:
    :arg timed_out:
    :arg took:
    :arg total:
    :arg updated:
    :arg version_conflicts:
    :arg throttled:
    :arg throttled_millis:
    :arg throttled_until:
    :arg throttled_until_millis:
    """

    _search: "UpdateByQueryBase[_R]"

    batches: int
    failures: Sequence["types.BulkIndexByScrollFailure"]
    noops: int
    deleted: int
    requests_per_second: float
    retries: "types.Retries"
    task: Union[str, int]
    timed_out: bool
    took: Any
    total: int
    updated: int
    version_conflicts: int
    throttled: Any
    throttled_millis: Any
    throttled_until: Any
    throttled_until_millis: Any

    def __init__(
        self,
        search: "Request[_R]",
        response: Dict[str, Any],
        doc_class: Optional[_R] = None,
    ):
        super(AttrDict, self).__setattr__("_search", search)
        super(AttrDict, self).__setattr__("_doc_class", doc_class)
        super().__init__(response)

    def success(self) -> bool:
        return not self.timed_out and not self.failures



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

from pytest import raises

from elasticsearch_dsl import function, query, utils


def test_empty_Q_is_match_all() -> None:
    q = query.Q()

    assert isinstance(q, query.MatchAll)
    assert query.MatchAll() == q


def test_combined_fields_to_dict() -> None:
    assert {
        "combined_fields": {
            "query": "this is a test",
            "fields": ["name", "body", "description"],
            "operator": "and",
        },
    } == query.CombinedFields(
        query="this is a test",
        fields=["name", "body", "description"],
        operator="and",
    ).to_dict()


def test_combined_fields_to_dict_extra() -> None:
    assert {
        "combined_fields": {
            "query": "this is a test",
            "fields": ["name", "body^2"],
            "operator": "or",
        },
    } == query.CombinedFields(
        query="this is a test",
        fields=["name", "body^2"],
        operator="or",
    ).to_dict()


def test_match_to_dict() -> None:
    assert {"match": {"f": "value"}} == query.Match(f="value").to_dict()


def test_match_to_dict_extra() -> None:
    assert {"match": {"f": "value", "boost": 2}} == query.Match(
        f="value", boost=2
    ).to_dict()


def test_fuzzy_to_dict() -> None:
    assert {"fuzzy": {"f": "value"}} == query.Fuzzy(f="value").to_dict()


def test_prefix_to_dict() -> None:
    assert {"prefix": {"f": "value"}} == query.Prefix(f="value").to_dict()


def test_term_to_dict() -> None:
    assert {"term": {"_type": "article"}} == query.Term(_type="article").to_dict()


def test_terms_to_dict() -> None:
    assert {"terms": {"_type": ["article", "section"]}} == query.Terms(
        _type=["article", "section"]
    ).to_dict()
    assert {"terms": {"_type": ["article", "section"], "boost": 1.1}} == query.Terms(
        _type=("article", "section"), boost=1.1
    ).to_dict()
    assert {"terms": {"_type": "article", "boost": 1.1}} == query.Terms(
        _type="article", boost=1.1
    ).to_dict()
    assert {
        "terms": {"_id": {"index": "my-other-index", "id": "my-id"}, "boost": 1.1}
    } == query.Terms(
        _id={"index": "my-other-index", "id": "my-id"}, boost=1.1
    ).to_dict()


def test_bool_to_dict() -> None:
    bool = query.Bool(must=[query.Match(f="value")], should=[])

    assert {"bool": {"must": [{"match": {"f": "value"}}]}} == bool.to_dict()


def test_dismax_to_dict() -> None:
    assert {"dis_max": {"queries": [{"term": {"_type": "article"}}]}} == query.DisMax(
        queries=[query.Term(_type="article")]
    ).to_dict()


def test_bool_from_dict_issue_318() -> None:
    d = {"bool": {"must_not": {"match": {"field": "value"}}}}
    q = query.Q(d)

    assert q == ~query.Match(field="value")


def test_repr() -> None:
    bool = query.Bool(must=[query.Match(f="value")], should=[])

    assert "Bool(must=[Match(f='value')])" == repr(bool)


def test_query_clone() -> None:
    bool = query.Bool(
        must=[query.Match(x=42)],
        should=[query.Match(g="v2")],
        must_not=[query.Match(title="value")],
    )
    bool_clone = bool._clone()

    assert bool == bool_clone
    assert bool is not bool_clone


def test_bool_converts_its_init_args_to_queries() -> None:
    q = query.Bool(must=[{"match": {"f": "value"}}])  # type: ignore

    assert len(q.must) == 1
    assert q.must[0] == query.Match(f="value")


def test_two_queries_make_a_bool() -> None:
    q1 = query.Match(f="value1")
    q2 = query.Match(message={"query": "this is a test", "opeartor": "and"})
    q = q1 & q2

    assert isinstance(q, query.Bool)
    assert [q1, q2] == q.must


def test_other_and_bool_appends_other_to_must() -> None:
    q1 = query.Match(f="value1")
    qb = query.Bool()

    q = q1 & qb
    assert q is not qb
    assert q.must[0] == q1


def test_bool_and_other_appends_other_to_must() -> None:
    q1 = query.Match(f="value1")
    qb = query.Bool()

    q = qb & q1
    assert q is not qb
    assert q.must[0] == q1


def test_bool_and_other_sets_min_should_match_if_needed() -> None:
    q1 = query.Q("term", category=1)
    q2 = query.Q(
        "bool", should=[query.Q("term", name="aaa"), query.Q("term", name="bbb")]
    )

    q = q1 & q2
    assert q == query.Bool(
        must=[q1],
        should=[query.Q("term", name="aaa"), query.Q("term", name="bbb")],
        minimum_should_match=1,
    )


def test_bool_with_different_minimum_should_match_should_not_be_combined() -> None:
    q1 = query.Q(
        "bool",
        minimum_should_match=2,
        should=[
            query.Q("term", field="aa1"),
            query.Q("term", field="aa2"),
            query.Q("term", field="aa3"),
            query.Q("term", field="aa4"),
        ],
    )
    q2 = query.Q(
        "bool",
        minimum_should_match=3,
        should=[
            query.Q("term", field="bb1"),
            query.Q("term", field="bb2"),
            query.Q("term", field="bb3"),
            query.Q("term", field="bb4"),
        ],
    )
    q3 = query.Q(
        "bool",
        minimum_should_match=4,
        should=[
            query.Q("term", field="cc1"),
            query.Q("term", field="cc2"),
            query.Q("term", field="cc3"),
            query.Q("term", field="cc4"),
        ],
    )

    q4 = q1 | q2
    assert q4 == query.Bool(should=[q1, q2])

    q5 = q1 | q2 | q3
    assert q5 == query.Bool(should=[q1, q2, q3])


def test_empty_bool_has_min_should_match_0() -> None:
    assert 0 == query.Bool()._min_should_match


def test_query_and_query_creates_bool() -> None:
    q1 = query.Match(f=42)
    q2 = query.Match(g=47)

    q = q1 & q2
    assert isinstance(q, query.Bool)
    assert q.must == [q1, q2]


def test_match_all_and_query_equals_other() -> None:
    q1 = query.Match(f=42)
    q2 = query.MatchAll()

    q = q1 & q2
    assert q1 == q


def test_not_match_all_is_match_none() -> None:
    q = query.MatchAll()

    assert ~q == query.MatchNone()


def test_not_match_none_is_match_all() -> None:
    q = query.MatchNone()

    assert ~q == query.MatchAll()


def test_invert_empty_bool_is_match_none() -> None:
    q = query.Bool()

    assert ~q == query.MatchNone()


def test_match_none_or_query_equals_query() -> None:
    q1 = query.Match(f=42)
    q2 = query.MatchNone()

    assert q1 | q2 == query.Match(f=42)


def test_match_none_and_query_equals_match_none() -> None:
    q1 = query.Match(f=42)
    q2 = query.MatchNone()

    assert q1 & q2 == query.MatchNone()


def test_bool_and_bool() -> None:
    qt1, qt2, qt3 = query.Match(f=1), query.Match(f=2), query.Match(f=3)

    q1 = query.Bool(must=[qt1], should=[qt2])
    q2 = query.Bool(must_not=[qt3])
    assert q1 & q2 == query.Bool(
        must=[qt1], must_not=[qt3], should=[qt2], minimum_should_match=0
    )

    q1 = query.Bool(must=[qt1], should=[qt1, qt2])
    q2 = query.Bool(should=[qt3])
    assert q1 & q2 == query.Bool(
        must=[qt1, qt3], should=[qt1, qt2], minimum_should_match=0
    )


def test_bool_and_bool_with_min_should_match() -> None:
    qt1, qt2 = query.Match(f=1), query.Match(f=2)
    q1 = query.Q("bool", minimum_should_match=1, should=[qt1])
    q2 = query.Q("bool", minimum_should_match=1, should=[qt2])

    assert query.Q("bool", must=[qt1, qt2]) == q1 & q2


def test_negative_min_should_match() -> None:
    qt1, qt2 = query.Match(f=1), query.Match(f=2)
    q1 = query.Q("bool", minimum_should_match=-2, should=[qt1])
    q2 = query.Q("bool", minimum_should_match=1, should=[qt2])

    with raises(ValueError):
        q1 & q2
    with raises(ValueError):
        q2 & q1


def test_percentage_min_should_match() -> None:
    qt1, qt2 = query.Match(f=1), query.Match(f=2)
    q1 = query.Q("bool", minimum_should_match="50%", should=[qt1])
    q2 = query.Q("bool", minimum_should_match=1, should=[qt2])

    with raises(ValueError):
        q1 & q2
    with raises(ValueError):
        q2 & q1


def test_inverted_query_becomes_bool_with_must_not() -> None:
    q = query.Match(f=42)

    assert ~q == query.Bool(must_not=[query.Match(f=42)])


def test_inverted_query_with_must_not_become_should() -> None:
    q = query.Q("bool", must_not=[query.Q("match", f=1), query.Q("match", f=2)])

    assert ~q == query.Q("bool", should=[query.Q("match", f=1), query.Q("match", f=2)])


def test_inverted_query_with_must_and_must_not() -> None:
    q = query.Q(
        "bool",
        must=[query.Q("match", f=3), query.Q("match", f=4)],
        must_not=[query.Q("match", f=1), query.Q("match", f=2)],
    )
    print((~q).to_dict())
    assert ~q == query.Q(
        "bool",
        should=[
            # negation of must
            query.Q("bool", must_not=[query.Q("match", f=3)]),
            query.Q("bool", must_not=[query.Q("match", f=4)]),
            # negation of must_not
            query.Q("match", f=1),
            query.Q("match", f=2),
        ],
    )


def test_double_invert_returns_original_query() -> None:
    q = query.Match(f=42)

    assert q == ~~q


def test_bool_query_gets_inverted_internally() -> None:
    q = query.Bool(must_not=[query.Match(f=42)], must=[query.Match(g="v")])

    assert ~q == query.Bool(
        should=[
            # negating must
            query.Bool(must_not=[query.Match(g="v")]),
            # negating must_not
            query.Match(f=42),
        ]
    )


def test_match_all_or_something_is_match_all() -> None:
    q1 = query.MatchAll()
    q2 = query.Match(f=42)

    assert (q1 | q2) == query.MatchAll()
    assert (q2 | q1) == query.MatchAll()


def test_or_produces_bool_with_should() -> None:
    q1 = query.Match(f=42)
    q2 = query.Match(g="v")

    q = q1 | q2
    assert q == query.Bool(should=[q1, q2])


def test_or_bool_doesnt_loop_infinitely_issue_37() -> None:
    q = query.Match(f=42) | ~query.Match(f=47)

    assert q == query.Bool(
        should=[query.Bool(must_not=[query.Match(f=47)]), query.Match(f=42)]
    )


def test_or_bool_doesnt_loop_infinitely_issue_96() -> None:
    q = ~query.Match(f=42) | ~query.Match(f=47)

    assert q == query.Bool(
        should=[
            query.Bool(must_not=[query.Match(f=42)]),
            query.Bool(must_not=[query.Match(f=47)]),
        ]
    )


def test_bool_will_append_another_query_with_or() -> None:
    qb = query.Bool(should=[query.Match(f="v"), query.Match(f="v2")])
    q = query.Match(g=42)

    assert (q | qb) == query.Bool(should=[query.Match(f="v"), query.Match(f="v2"), q])


def test_bool_queries_with_only_should_get_concatenated() -> None:
    q1 = query.Bool(should=[query.Match(f=1), query.Match(f=2)])
    q2 = query.Bool(should=[query.Match(f=3), query.Match(f=4)])

    assert (q1 | q2) == query.Bool(
        should=[query.Match(f=1), query.Match(f=2), query.Match(f=3), query.Match(f=4)]
    )


def test_two_bool_queries_append_one_to_should_if_possible() -> None:
    q1 = query.Bool(should=[query.Match(f="v")])
    q2 = query.Bool(must=[query.Match(f="v")])

    assert (q1 | q2) == query.Bool(
        should=[query.Match(f="v"), query.Bool(must=[query.Match(f="v")])]
    )
    assert (q2 | q1) == query.Bool(
        should=[query.Match(f="v"), query.Bool(must=[query.Match(f="v")])]
    )


def test_queries_are_registered() -> None:
    assert "match" in query.Query._classes
    assert query.Query._classes["match"] is query.Match


def test_defining_query_registers_it() -> None:
    class MyQuery(query.Query):
        name = "my_query"

    assert "my_query" in query.Query._classes
    assert query.Query._classes["my_query"] is MyQuery


def test_Q_passes_query_through() -> None:
    q = query.Match(f="value1")

    assert query.Q(q) is q


def test_Q_constructs_query_by_name() -> None:
    q = query.Q("match", f="value")

    assert isinstance(q, query.Match)
    assert {"f": "value"} == q._params


def test_Q_translates_double_underscore_to_dots_in_param_names() -> None:
    q = query.Q("match", comment__author="honza")

    assert {"comment.author": "honza"} == q._params


def test_Q_doesn_translate_double_underscore_to_dots_in_param_names() -> None:
    q = query.Q("match", comment__author="honza", _expand__to_dot=False)

    assert {"comment__author": "honza"} == q._params


def test_Q_constructs_simple_query_from_dict() -> None:
    q = query.Q({"match": {"f": "value"}})

    assert isinstance(q, query.Match)
    assert {"f": "value"} == q._params


def test_Q_constructs_compound_query_from_dict() -> None:
    q = query.Q({"bool": {"must": [{"match": {"f": "value"}}]}})

    assert q == query.Bool(must=[query.Match(f="value")])


def test_Q_raises_error_when_passed_in_dict_and_params() -> None:
    with raises(Exception):
        # Ignore types as it's not a valid call
        query.Q({"match": {"f": "value"}}, f="value")  # type: ignore[call-overload]


def test_Q_raises_error_when_passed_in_query_and_params() -> None:
    q = query.Match(f="value1")

    with raises(Exception):
        # Ignore types as it's not a valid call signature
        query.Q(q, f="value")  # type: ignore[call-overload]


def test_Q_raises_error_on_unknown_query() -> None:
    with raises(Exception):
        query.Q("not a query", f="value")


def test_match_all_and_anything_is_anything() -> None:
    q = query.MatchAll()

    s = query.Match(f=42)
    assert q & s == s
    assert s & q == s


def test_function_score_with_functions() -> None:
    q = query.Q(
        "function_score",
        functions=[query.SF("script_score", script="doc['comment_count'] * _score")],
    )

    assert {
        "function_score": {
            "functions": [{"script_score": {"script": "doc['comment_count'] * _score"}}]
        }
    } == q.to_dict()


def test_function_score_with_no_function_is_boost_factor() -> None:
    q = query.Q(
        "function_score",
        functions=[query.SF({"weight": 20, "filter": query.Q("term", f=42)})],
    )

    assert {
        "function_score": {"functions": [{"filter": {"term": {"f": 42}}, "weight": 20}]}
    } == q.to_dict()


def test_function_score_to_dict() -> None:
    q = query.Q(
        "function_score",
        query=query.Q("match", title="python"),
        functions=[
            query.SF("random_score"),
            query.SF(
                "field_value_factor",
                field="comment_count",
                filter=query.Q("term", tags="python"),
            ),
        ],
    )

    d = {
        "function_score": {
            "query": {"match": {"title": "python"}},
            "functions": [
                {"random_score": {}},
                {
                    "filter": {"term": {"tags": "python"}},
                    "field_value_factor": {"field": "comment_count"},
                },
            ],
        }
    }
    assert d == q.to_dict()


def test_function_score_class_based_to_dict() -> None:
    q = query.FunctionScore(
        query=query.Match(title="python"),
        functions=[
            function.RandomScore(),
            function.FieldValueFactor(
                field="comment_count",
                filter=query.Term(tags="python"),
            ),
        ],
    )

    d = {
        "function_score": {
            "query": {"match": {"title": "python"}},
            "functions": [
                {"random_score": {}},
                {
                    "filter": {"term": {"tags": "python"}},
                    "field_value_factor": {"field": "comment_count"},
                },
            ],
        }
    }
    assert d == q.to_dict()


def test_function_score_with_single_function() -> None:
    d = {
        "function_score": {
            "filter": {"term": {"tags": "python"}},
            "script_score": {"script": "doc['comment_count'] * _score"},
        }
    }

    q = query.Q(d)
    assert isinstance(q, query.FunctionScore)
    assert isinstance(q.filter, query.Term)
    assert len(q.functions) == 1

    sf = q.functions[0]
    assert isinstance(sf, function.ScriptScore)
    assert "doc['comment_count'] * _score" == sf.script


def test_function_score_from_dict() -> None:
    d = {
        "function_score": {
            "filter": {"term": {"tags": "python"}},
            "functions": [
                {
                    "filter": {"terms": {"tags": "python"}},
                    "script_score": {"script": "doc['comment_count'] * _score"},
                },
                {"boost_factor": 6},
            ],
        }
    }

    q = query.Q(d)
    assert isinstance(q, query.FunctionScore)
    assert isinstance(q.filter, query.Term)
    assert len(q.functions) == 2

    sf = q.functions[0]
    assert isinstance(sf, function.ScriptScore)
    assert isinstance(sf.filter, query.Terms)

    sf = q.functions[1]
    assert isinstance(sf, function.BoostFactor)
    assert 6 == sf.value
    assert {"boost_factor": 6} == sf.to_dict()


def test_script_score() -> None:
    d = {
        "script_score": {
            "query": {"match_all": {}},
            "script": {"source": "...", "params": {}},
        }
    }
    q = query.Q(d)

    assert isinstance(q, query.ScriptScore)
    assert isinstance(q.query, query.MatchAll)
    assert q.script == {"source": "...", "params": {}}
    assert q.to_dict() == d


def test_expand_double_underscore_to_dot_setting() -> None:
    q = query.Term(comment__count=2)
    assert q.to_dict() == {"term": {"comment.count": 2}}
    utils.EXPAND__TO_DOT = False
    q = query.Term(comment__count=2)
    assert q.to_dict() == {"term": {"comment__count": 2}}
    utils.EXPAND__TO_DOT = True


def test_knn_query() -> None:
    q = query.Knn(field="image-vector", query_vector=[-5, 9, -12], num_candidates=10)
    assert q.to_dict() == {
        "knn": {
            "field": "image-vector",
            "query_vector": [-5, 9, -12],
            "num_candidates": 10,
        }
    }
